import os
from typing import Dict, List, Optional, Set, Tuple, Union
from uuid import UUID

import magic

from supervisely import (
    Api,
    PointcloudEpisodeAnnotation,
    batched,
    generate_free_name,
    is_development,
    logger,
)
from supervisely.api.module_api import ApiField
from supervisely.convert.base_converter import BaseConverter
from supervisely.io.fs import get_file_ext, get_file_name_with_ext
from supervisely.io.json import load_json_file
from supervisely.pointcloud.pointcloud import ALLOWED_POINTCLOUD_EXTENSIONS
from supervisely.pointcloud.pointcloud import validate_ext as validate_pcd_ext
from supervisely.pointcloud_annotation.constants import OBJECT_KEY
from supervisely.project.project_settings import LabelingInterface
from supervisely.video_annotation.key_id_map import KeyIdMap


class PointcloudEpisodeConverter(BaseConverter):
    """Base converter for pointcloud episodes (frame-indexed pointclouds + a shared episode annotation)."""

    allowed_exts = ALLOWED_POINTCLOUD_EXTENSIONS
    modality = "pointcloud episodes"

    class Item(BaseConverter.BaseItem):
        """Base pointcloud-episode item with frame index and optional related images bundle."""

        def __init__(
            self,
            item_path,
            frame_number: int,
            ann_data: Optional[str] = None,
            related_images: Optional[list] = None,
            custom_data: Optional[dict] = None,
        ):
            """
            :param item_path: Path to frame pointcloud.
            :type item_path: str
            :param frame_number: Frame index in episode.
            :type frame_number: int
            :param ann_data: Annotation path.
            :type ann_data: str, optional
            :param related_images: Related images list.
            :type related_images: list, optional
            :param custom_data: Extra data.
            :type custom_data: dict, optional
            """
            self._name: str = None
            self._path = item_path
            self._frame_number = frame_number
            self._ann_data = ann_data
            self._related_images = related_images if related_images is not None else []
            self._type = "point_cloud_episode"
            self._custom_data = custom_data if custom_data is not None else {}

        @property
        def frame_number(self) -> int:
            return self._frame_number

        def create_empty_annotation(self) -> PointcloudEpisodeAnnotation:
            return PointcloudEpisodeAnnotation()

        def set_related_images(self, related_images: Tuple[str, str, Optional[str]]) -> None:
            """Adds related image to the item.

            related_images tuple:
                - path to image
                - path to .json with image metadata
                - path to .figures.json (can be None if no figures)
            """
            self._related_images.append(related_images)

    def __init__(
        self,
        input_data: str,
        labeling_interface: Optional[Union[LabelingInterface, str]],
        upload_as_links: bool = False,
        remote_files_map: Optional[Dict[str, str]] = None,
    ):
        """See :class:`~supervisely.convert.base_converter.BaseConverter` for params."""
        super().__init__(input_data, labeling_interface, upload_as_links, remote_files_map)
        self._annotation = None
        self._frame_pointcloud_map = None
        self._frame_count = None

    @property
    def format(self):
        return self._converter.format

    @property
    def frame_count(self):
        if self._frame_count is None:
            self._frame_count = len(self._items)
        return self._frame_count

    @property
    def ann_ext(self):
        return None

    @property
    def key_file_ext(self):
        return None

    @staticmethod
    def validate_ann_file(ann_path, meta=None):
        return False

    def upload_dataset(
        self,
        api: Api,
        dataset_id: int,
        batch_size: int = 1,
        log_progress=True,
    ):
        """Upload converted data to Supervisely"""

        meta, renamed_classes, renamed_tags = self.merge_metas_with_conflicts(api, dataset_id)

        existing_pcde_infos = api.pointcloud_episode.get_list(dataset_id)
        existing_names = set([pcde.name for pcde in existing_pcde_infos])

        if log_progress:
            progress, progress_cb = self.get_progress(
                self.items_count, "Uploading pointcloud episodes..."
            )
        else:
            progress_cb = None

        frame_to_pointcloud_ids: Dict[int, int] = {}
        pcl_to_rimg_figures: Dict[int, Dict[str, List[Dict]]] = {}
        pcl_to_hash_to_id: Dict[int, Dict[str, int]] = {}
        used_related_image_names: Set[str] = set()
        try:
            existing_pcde_ids = [pcde.id for pcde in existing_pcde_infos]
            for pcde_ids_batch in batched(existing_pcde_ids, batch_size=200):
                related_images = api.pointcloud.get_list_related_images_batch(
                    dataset_id, pcde_ids_batch
                )
                for related_image in related_images:
                    related_image_name = related_image.get(ApiField.NAME)
                    if related_image_name is not None:
                        used_related_image_names.add(related_image_name)
        except Exception as e:
            logger.debug(
                f"Failed to fetch existing related image names for dataset ID:{dataset_id}: {repr(e)}"
            )
        key_id_map = KeyIdMap()
        for batch in batched(self._items, batch_size=batch_size):
            item_names = []
            item_paths = []
            item_metas = []
            for item in batch:
                item.name = generate_free_name(
                    existing_names, item.name, with_ext=True, extend_used_names=True
                )
                item_names.append(item.name)
                item_paths.append(item.path)

            pcd_infos = api.pointcloud_episode.upload_paths(
                dataset_id,
                item_names,
                item_paths,
            )
            pcd_ids = [pcd_info.id for pcd_info in pcd_infos]

            for item, pcd_id in zip(batch, pcd_ids):
                frame_to_pointcloud_ids[item._frame_number] = pcd_id
                item_metas.append({"frame": item._frame_number})
                rimg_infos = []
                camera_names = []
                if len(item._related_images) > 0:
                    img_paths: List[str] = []
                    ann_paths: List[str] = []
                    fig_paths: List[Optional[str]] = []

                    for triple in item._related_images:
                        img_paths.append(triple[0])
                        ann_paths.append(triple[1])
                        fig_paths.append(triple[2] if len(triple) > 2 else None)

                    rimg_hashes = api.pointcloud_episode.upload_related_images(img_paths)

                    for img_ind, (img_hash, ann_path, fig_path) in enumerate(
                        zip(rimg_hashes, ann_paths, fig_paths)
                    ):
                        meta_json = load_json_file(ann_path)
                        try:
                            if ApiField.META not in meta_json:
                                raise ValueError("Related image meta not found in json file.")
                            if ApiField.NAME not in meta_json:
                                raise ValueError("Related image name not found in json file.")
                            if "deviceId" not in meta_json[ApiField.META].keys():
                                camera_names.append(f"CAM_{str(img_ind).zfill(2)}")
                            else:
                                camera_names.append(meta_json[ApiField.META]["deviceId"])
                            related_image_name = generate_free_name(
                                used_related_image_names,
                                meta_json[ApiField.NAME],
                                with_ext=True,
                                extend_used_names=True,
                            )
                            rimg_infos.append(
                                {
                                    ApiField.ENTITY_ID: pcd_id,
                                    ApiField.NAME: related_image_name,
                                    ApiField.HASH: img_hash,
                                    ApiField.META: meta_json[ApiField.META],
                                }
                            )

                            if fig_path is not None:
                                try:
                                    figs_json = load_json_file(fig_path)
                                    pcl_to_rimg_figures.setdefault(pcd_id, {})[img_hash] = figs_json
                                except Exception as e:
                                    logger.debug(
                                        f"Failed to read figures json '{fig_path}': {repr(e)}"
                                    )

                        except Exception as e:
                            logger.warning(
                                f"Failed to process related image meta '{ann_path}': {repr(e)}"
                            )
                            continue

                    uploaded_rimgs = api.pointcloud.add_related_images(rimg_infos, camera_names)
                    try:
                        # build mapping hash->id
                        for info, uploaded in zip(rimg_infos, uploaded_rimgs):
                            img_hash = info.get(ApiField.HASH)
                            img_id = (
                                uploaded.get(ApiField.ID)
                                if isinstance(uploaded, dict)
                                else getattr(uploaded, "id", None)
                            )
                            if img_hash is not None and img_id is not None:
                                pcl_to_hash_to_id.setdefault(pcd_id, {})[img_hash] = img_id
                    except Exception as e:
                        logger.debug(
                            f"Failed to build hash->ID mapping for related images: {repr(e)}"
                        )

            if log_progress:
                progress_cb(len(batch))

        if self.items_count > 0:
            ann = self.to_supervisely(self._items[0], meta, renamed_classes, renamed_tags)
            api.pointcloud_episode.annotation.append(
                dataset_id, ann, frame_to_pointcloud_ids, key_id_map
            )

        # ---- upload figures for processed batch ----
        if len(pcl_to_rimg_figures) > 0:
            if log_progress:
                progress, progress_cb = self.get_progress(
                    self.items_count, "Uploading figures for related images..."
                )
            else:
                progress_cb = None

            try:
                dataset_info = api.dataset.get_info_by_id(dataset_id)
                project_id = dataset_info.project_id

                figures_to_upload: List[Dict] = []
                for pcl_id, hash_to_figs in pcl_to_rimg_figures.items():
                    hash_to_ids = pcl_to_hash_to_id.get(pcl_id, {})
                    if len(hash_to_ids) == 0:
                        continue

                    for img_hash, figs_json in hash_to_figs.items():
                        if img_hash not in hash_to_ids:
                            continue
                        img_id = hash_to_ids[img_hash]

                        for fig in figs_json:
                            try:
                                fig[ApiField.ENTITY_ID] = img_id
                                fig[ApiField.DATASET_ID] = dataset_id
                                fig[ApiField.PROJECT_ID] = project_id
                                if OBJECT_KEY in fig:
                                    fig[ApiField.OBJECT_ID] = key_id_map.get_object_id(
                                        UUID(fig[OBJECT_KEY])
                                    )
                            except Exception as e:
                                logger.debug(
                                    f"Failed to process figure json for img_hash={img_hash}: {repr(e)}"
                                )
                                continue

                        figures_to_upload.extend(figs_json)

                if len(figures_to_upload) > 0:
                    try:
                        api.image.figure.create_bulk(
                            figures_json=figures_to_upload, dataset_id=dataset_id
                        )
                    except Exception as e:
                        logger.debug(f"Failed to upload figures for related images: {repr(e)}")

                if log_progress:
                    progress_cb(len(pcl_to_rimg_figures))

            except Exception as e:
                logger.debug(f"Unexpected error during figures upload: {repr(e)}")

        if log_progress:
            if is_development():
                progress.close()
        logger.info(f"Dataset ID:{dataset_id} has been successfully uploaded.")

    def _collect_items_if_format_not_detected(self) -> Tuple[List[Item], bool, Set[str]]:
        only_modality_items = True
        unsupported_exts = set()
        pcd_dict = {}
        frames_pcd_map = None
        rimg_dict, rimg_json_dict, rimg_fig_dict = {}, {}, {}
        for root, _, files in os.walk(self._input_data):
            dir_name = os.path.basename(root)
            for file in files:
                full_path = os.path.join(root, file)
                if file == "frame_pointcloud_map.json":
                    frames_pcd_map = load_json_file(full_path)
                    continue

                ext = get_file_ext(full_path)
                recognized_ext = self._get_image_subtype(full_path)
                if file.endswith(".figures.json"):
                    rimg_fig_dict[file] = full_path
                elif ext == ".json":
                    rimg_json_dict[file] = full_path
                elif recognized_ext:
                    if ext.lower() == ".pcd":
                        logger.warning(
                            f"File '{file}' has been recognized as '.{recognized_ext}' format. Skipping."
                        )
                        continue
                    if dir_name not in rimg_dict:
                        rimg_dict[dir_name] = []
                    rimg_dict[dir_name].append(full_path)
                elif ext.lower() in self.allowed_exts:
                    try:
                        validate_pcd_ext(ext)
                        pcd_dict[file] = full_path
                    except:
                        pass
                else:
                    only_modality_items = False
                    unsupported_exts.add(ext)

        items = []
        updated_frames_pcd_map = {}
        if frames_pcd_map:
            list_of_pcd_names = list(frames_pcd_map.values())
        else:
            list_of_pcd_names = sorted(pcd_dict.keys())

        for i, pcd_name in enumerate(list_of_pcd_names):
            if pcd_name in pcd_dict:
                updated_frames_pcd_map[i] = pcd_name
                item = self.Item(pcd_dict[pcd_name], i)
                rimg_dir_name = pcd_name.replace(".pcd", "_pcd")
                rimgs = rimg_dict.get(rimg_dir_name, [])
                for rimg_path in rimgs:
                    rimg_ann_name = f"{get_file_name_with_ext(rimg_path)}.json"
                    if rimg_ann_name in rimg_json_dict:
                        rimg_ann_path = rimg_json_dict[rimg_ann_name]
                        rimg_fig_name = f"{get_file_name_with_ext(rimg_path)}.figures.json"
                        rimg_fig_path = rimg_fig_dict.get(rimg_fig_name, None)
                        if rimg_fig_path is not None and not os.path.exists(rimg_fig_path):
                            rimg_fig_path = None
                        item.set_related_images((rimg_path, rimg_ann_path, rimg_fig_path))
                items.append(item)
            else:
                logger.warning(f"Pointcloud file {pcd_name} not found. Skipping frame.")
                continue
        self._frame_pointcloud_map = updated_frames_pcd_map
        self._frame_count = len(self._frame_pointcloud_map)

        return items, only_modality_items, unsupported_exts

    @staticmethod
    def _get_image_subtype(path: str) -> Optional[str]:
        try:
            mime_type = magic.from_file(path, mime=True)
        except Exception:
            return None
        if mime_type.startswith("image/"):
            return mime_type.split("/", 1)[1]
        return None



from __future__ import annotations

from typing import List
from supervisely.app import DataJson, StateJson
from supervisely.app.widgets import Widget


"""

l = sly.app.widgets.Text(text="left part", status="success")
items = [
    sly.app.widgets.Select.Item(label="CPU", value="cpu"),
    sly.app.widgets.Select.Item(label="GPU 0", value="cuda:0"),
    sly.app.widgets.Select.Item(value="option3"),
]
r = sly.app.widgets.Select(items=items, filterable=True, placeholder="select me")

# sidebar = sly.app.widgets.Sidebar(left_pane=l, right_pane=item)

g1_items = [
    sly.app.widgets.Menu.Item(title="m1", content=r),
    sly.app.widgets.Menu.Item(title="m2", content=l),
]
g2_items = [
    sly.app.widgets.Menu.Item(title="m3"),
    sly.app.widgets.Menu.Item(title="m4"),
]
g1 = sly.app.widgets.Menu.Group("g1", g1_items)
g2 = sly.app.widgets.Menu.Group("g2", g2_items)
menu = sly.app.widgets.Menu(groups=[g1, g2])
# menu = sly.app.widgets.Menu(items=g1_items)

"""


class Menu(Widget):
    """Sidebar menu widget with items optionally grouped into sections."""

    class Item:
        """Single menu item descriptor (title + optional icon and content widget)."""

        def __init__(
            self,
            title: str,
            # heading: str = None,
            content: Widget = None,
            index: str = None,
            icon: str = None,
        ) -> Menu.Item:
            """
            :param title: Item title.
            :type title: str
            :param content: Optional widget to show when item is selected.
            :type content: Widget, optional
            :param index: Unique index for routing. Defaults to title.
            :type index: str, optional
            :param icon: Icon class (e.g. zmdi).
            :type icon: str, optional
            """
            self.title = title
            # self.heading = title if heading is None else heading
            self.index = index
            self.content = content
            if index is None:
                self.index = str(self.title)
            self.icon = icon

        def to_json(self):
            return {
                "title": self.title,
                # "heading": self.heading,
                "index": self.index,
                "icon": self.icon,
            }

    class Group:
        """Group of menu items displayed under a section title."""

        def __init__(self, title: str, items: List[Menu.Item] = None) -> Menu.Group:
            """
            :param title: Group section title.
            :type title: str
            :param items: List of Menu.Item in this group.
            :type items: List[Menu.Item], optional
            """
            self.title = title
            self.items = items

        def to_json(self):
            return {
                "title": self.title,
                "items": [item.to_json() for item in self.items],
            }

    def __init__(
        self,
        items: List[Menu.Item] = None,
        groups: List[Menu.Group] = None,
        index: str = None,
        width_percent: int = 25,
        widget_id: str = None,
    ):
        """
        :param items: Flat list of menu items. Mutually exclusive with groups.
        :type items: List[Menu.Item], optional
        :param groups: List of Menu.Group. Mutually exclusive with items.
        :type groups: List[Menu.Group], optional
        :param index: Initially selected item index.
        :type index: str, optional
        :param width_percent: Sidebar width as percent (0–100).
        :type width_percent: int
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional

        :raises ValueError: If both items and groups are None, or both provided.
        """
        if items is None and groups is None:
            raise ValueError("One of the arguments has to be defined: items or groups")

        if items is not None and groups is not None:
            raise ValueError(
                "Only one of the arguments has to be defined: items or groups"
            )

        self._items = items
        if self._items is None:
            self._items = []
        self._groups = groups
        if self._groups is None:
            self._groups = []
        self._width_percent = width_percent
        self._options = {"sidebarWidth": self._width_percent}
        self._index = index
        if self._index is None:
            self._index = self._get_default_index()
        else:
            self._validate_index()

        StateJson()["app_body_padding"] = "0px"
        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        res = {"items": [], "groups": []}
        if self._items is not None:
            res["items"] = [item.to_json() for item in self._items]
        if self._groups is not None:
            res["groups"] = [g.to_json() for g in self._groups]
        return res

    def _validate_index(self):
        for item in self._items:
            if self._index == item.index:
                return

        for group in self._groups:
            for item in group.items:
                if self._index == item.index:
                    return

        raise ValueError(
            f"Index {self._index} not found, there is no item in menu with such index"
        )

    def _get_default_index(self):
        if len(self._items) > 0:
            return self._items[0].index
        if len(self._groups) > 0:
            return self._groups[0].items[0].index

    def get_json_state(self):
        return {"menuIndex": self._index}



from __future__ import annotations

import threading
import time
import traceback
import uuid
from concurrent.futures import Future, ThreadPoolExecutor
from functools import partial
from typing import Any, Dict, List, Tuple, Union

from supervisely.nn.utils import get_gpu_usage, get_ram_usage
from supervisely.sly_logger import logger
from supervisely.task.progress import Progress


class InferenceRequest:
    """State container for a single inference run (progress, results queue, lifecycle stage and cancellation)."""

    PAUSE_SLEEP_INTERVAL = 1.0
    PAUSE_SLEEP_MAX_WAIT = 60 * 60  # 1 hour
    PENDING_RESULTS_MAX_SIZE = 500

    class Stage:
        """Human-readable stage labels used for progress reporting."""

        PREPARING = "Preparing model for inference..."
        INFERENCE = "Running inference..."
        FINISHED = "Finished"
        CANCELLED = "Cancelled"
        ERROR = "Error"

    def __init__(
        self,
        uuid_: str = None,
        ttl: Union[int, None] = 60 * 60,
        manager: InferenceRequestsManager = None,
    ):
        """
        :param uuid_: Request ID.
        :type uuid_: str
        :param ttl: Time-to-live in seconds.
        :type ttl: Union[int, None]
        :param manager: Optional InferenceRequestsManager.
        :type manager: InferenceRequestsManager
        """
        if uuid_ is None:
            uuid_ = uuid.uuid5(namespace=uuid.NAMESPACE_URL, name=f"{time.time()}").hex
        self._uuid = uuid_
        self._ttl = ttl
        self.manager = manager
        self.context = {}
        self._lock = threading.Lock()
        self._stage = InferenceRequest.Stage.PREPARING
        self._pending_results = []
        self._final_result = None
        self._exception = None
        self._stopped = threading.Event()
        self._paused_time = None
        self._paused = threading.Event()
        self._progress_log_interval = 5.0
        self._last_progress_report_time = 0
        self.progress = Progress(
            message=self._stage,
            total_cnt=1,
            need_info_log=True,
            update_task_progress=False,
            log_extra={"inference_request_uuid": self._uuid},
        )
        self._created_at = time.monotonic()
        self._updated_at = self._created_at
        self._finished = False

        self.tracker = None

        self.global_progress = None
        self.global_progress_total = 1
        self.global_progress_current = 0

    def _updated(self):
        self._updated_at = time.monotonic()

    @property
    def uuid(self):
        return self._uuid

    @property
    def created_at(self):
        return self._created_at

    @property
    def updated_at(self):
        return self._updated_at

    @property
    def stage(self):
        return self._stage

    @property
    def final_result(self):
        with self._lock:
            return self._final_result

    @final_result.setter
    def final_result(self, result: Any):
        with self._lock:
            self._final_result = result
            self._updated()

    def add_results(self, results: List[Dict]):
        with self._lock:
            self._pending_results.extend(results)
            if self.pending_num() >= self.PENDING_RESULTS_MAX_SIZE:
                logger.warning(
                    f"Pending results exceeded max size {self.PENDING_RESULTS_MAX_SIZE}, pausing inference request {self._uuid} for {self.PAUSE_SLEEP_MAX_WAIT} seconds."
                )
                self.pause()
            self._updated()

    def pop_pending_results(self, n: int = None):
        with self._lock:
            self.continue_if_paused()
            if len(self._pending_results) == 0:
                return []
            if n is None:
                n = len(self._pending_results)
            if n > len(self._pending_results):
                n = len(self._pending_results)
            results = self._pending_results[:n]
            self._pending_results = self._pending_results[n:]
            self._updated()
            return results

    def pending_num(self):
        return len(self._pending_results)

    def set_stage(self, stage: str, current: int = None, total: int = None, is_size: bool = False):
        with self._lock:
            self._stage = stage
            self.progress.message = self._stage
            if current is not None:
                self.progress.current = current
            if total is not None:
                logger.debug("setting total = %s", total)
                self.progress.total = total
            if is_size:
                self.progress.is_size = True
            self.progress._refresh_labels()
            self.progress.report_progress()
            if self._stage == InferenceRequest.Stage.INFERENCE:
                self.global_progress_total = total
                self.global_progress_current = current
                self.manager.global_progress.inference_started(
                    current=self.global_progress_current,
                    total=self.global_progress_total,
                )
            self._updated()

    def set_progress_log_interval(self, interval: float):
        self._progress_log_interval = interval

    def done(self, n=1):
        with self._lock:
            if (
                self._progress_log_interval is None
                or time.monotonic() - self._last_progress_report_time > self._progress_log_interval
            ) or (self.progress.current + n >= self.progress.total):
                self.progress.iters_done_report(n)
                self._last_progress_report_time = time.monotonic()
            else:
                self.progress.iters_done(n)
            if self._stage == InferenceRequest.Stage.INFERENCE:
                self.global_progress_current += n
                if self.manager is not None:
                    self.manager.done(n)
            self._updated()

    @property
    def exception(self):
        return self._exception

    @exception.setter
    def exception(self, exc: Exception):
        self._exception = exc
        self.set_stage(InferenceRequest.Stage.ERROR)
        self._updated()

    def is_inferring(self):
        return self.stage == InferenceRequest.Stage.INFERENCE

    def stop(self):
        self._stopped.set()
        self._updated()

    def is_stopped(self):
        return self._stopped.is_set()

    def pause(self):
        self._paused.set()
        self._paused_time = time.monotonic()
        self._updated()

    def is_paused(self):
        return self._paused.is_set()

    def continue_if_paused(self):
        if not self.is_paused():
            return
        logger.debug(f"Continuing paused inference request {self._uuid}")
        self._paused.clear()
        self._paused_time = None
        self._updated()

    def paused_for(self):
        if not self.is_paused():
            return 0
        return time.monotonic() - self._paused_time

    def is_finished(self):
        return self._finished

    def is_expired(self):
        if self._ttl is None:
            return False
        return time.monotonic() - self._updated_at > self._ttl

    def progress_json(self):
        return {
            "message": self.progress.message,
            "status": self.progress.message,
            "current": self.progress.current,
            "total": self.progress.total,
            "is_size": self.progress.is_size,
        }

    def exception_json(self):
        if self._exception is None:
            return None
        return {
            "type": type(self._exception).__name__,
            "message": str(self._exception),
            "traceback": str(traceback.format_exc()),
        }

    def to_json(self):
        return {
            "uuid": self._uuid,
            "stage": str(self._stage),
            "progress": self.progress_json(),
            "pending_results": self.pending_num(),
            "final_result": self._final_result is not None,
            "exception": self.exception_json(),
            "is_inferring": self.is_inferring(),
            "stopped": self.is_stopped(),
            "paused": self.is_paused(),
            "finished": self._finished,
            "created_at": self._created_at,
            "updated_at": self._updated_at,
        }

    def on_inference_end(self):
        if self._stage not in (
            InferenceRequest.Stage.FINISHED,
            InferenceRequest.Stage.CANCELLED,
            InferenceRequest.Stage.ERROR,
        ):
            if self.is_stopped():
                self.set_stage(InferenceRequest.Stage.CANCELLED)
            else:
                self.set_stage(InferenceRequest.Stage.FINISHED)
        self._finished = True
        self._updated()

    def get_usage(self):
        ram_allocated, ram_total = get_ram_usage()
        gpu_allocated, gpu_total = get_gpu_usage()
        return {
            "gpu_memory": {
                "allocated": gpu_allocated,
                "total": gpu_total,
            },
            "ram_memory": {
                "allocated": ram_allocated,
                "total": ram_total,
            },
        }

    def status(self):
        status_data = self.to_json()
        for key in ["pending_results", "final_result", "created_at", "updated_at"]:
            status_data.pop(key, None)
        status_data.update(self.get_usage())
        return status_data


class GlobalProgress:
    """Aggregated progress indicator for all inference requests managed by a server instance."""

    def __init__(self):
        self.progress = Progress(message="Ready", total_cnt=1)
        self._lock = threading.Lock()

    def set_message(self, message: str):
        with self._lock:
            if self.progress.message != message:
                self.progress.message = message
                self.progress.report_progress()

    def set_ready(self):
        with self._lock:
            self.progress.message = "Ready"
            self.progress.current = 0
            self.progress.total = 1
            self.progress.report_progress()

    def done(self, n=1):
        with self._lock:
            self.progress.iters_done_report(n)
        if self.progress.current >= self.progress.total:
            self.set_ready()

    def inference_started(self, current: int, total: int):
        with self._lock:
            if self.progress.message == "Ready":
                self.progress.total = total
                self.progress.current = current
            else:
                self.progress.total += total
                self.progress.current += current
        self.set_message("Inferring model...")

    def inference_finished(self, current: int, total: int):
        with self._lock:
            if self.progress.message == "Ready":
                return
            self.progress.current = self.progress.current - current
            self.progress.total = self.progress.total - total
        if self.progress.current >= self.progress.total:
            self.set_ready()

    def to_json(self):
        return {
            "message": self.progress.message,
            "current": self.progress.current,
            "total": self.progress.total,
        }


class InferenceRequestsManager:
    """Thread-safe registry and monitor for active :class:`InferenceRequest` instances."""

    def __init__(self, executor: ThreadPoolExecutor = None):
        """
        :param executor: Thread pool for request execution.
        :type executor: ThreadPoolExecutor
        """
        if executor is None:
            executor = ThreadPoolExecutor(max_workers=1)
        self._executor = executor
        self._inference_requests: Dict[str, InferenceRequest] = {}
        self._lock = threading.Lock()
        self._stop_event = threading.Event()
        self._monitor_thread = threading.Thread(target=self.monitor, daemon=True)
        self._monitor_thread.start()
        self.global_progress = GlobalProgress()

    def __del__(self):
        try:
            self._executor.shutdown(wait=False)
            self._stop_event.set()
            self._monitor_thread.join(timeout=5)
        finally:
            logger.debug("InferenceRequestsManager was deleted")

    def add(self, inference_request: InferenceRequest):
        with self._lock:
            self._inference_requests[inference_request.uuid] = inference_request

    def remove(self, inference_request_uuid: str):
        with self._lock:
            inference_request = self._inference_requests.get(inference_request_uuid)
            if inference_request is not None:
                inference_request.stop()
                del self._inference_requests[inference_request_uuid]

    def remove_after(self, inference_request_uuid, wait_time=0):
        with self._lock:
            inference_request = self._inference_requests.get(inference_request_uuid)
            if inference_request is not None:
                inference_request.stop()
                inference_request._ttl = wait_time
                inference_request._updated()

    def get(self, inference_request_uuid: str):
        if inference_request_uuid is None:
            return None
        try:
            return self._inference_requests[inference_request_uuid]
        except Exception as ex:
            raise RuntimeError(
                f"inference_request_uuid {inference_request_uuid} was given, "
                f"but there is no such uuid in 'self._inference_requests' ({len(self._inference_requests)} items)"
            ) from ex

    def create(self, inference_request_uuid: str = None) -> InferenceRequest:
        inference_request = InferenceRequest(uuid_=inference_request_uuid, manager=self)
        self.add(inference_request)
        return inference_request

    def monitor(self):
        while self._stop_event.is_set() is False:
            for inference_request_uuid in list(self._inference_requests.keys()):
                inference_request = self._inference_requests.get(inference_request_uuid)
                if inference_request is None:
                    continue
                if (
                    inference_request.is_paused()
                    and inference_request.paused_for() < InferenceRequest.PAUSE_SLEEP_MAX_WAIT
                ):
                    continue
                if inference_request.is_expired():
                    self.remove(inference_request_uuid)
                    logger.debug(f"Expired inference request {inference_request_uuid} was deleted")
            time.sleep(30)

    def done(self, n=1):
        with self._lock:
            self.global_progress.done(n)

    def _on_inference_start(self, inference_request: InferenceRequest):
        if inference_request.uuid not in self._inference_requests:
            self.add(inference_request)

    def _on_inference_end(self, future, inference_request_uuid: str):
        logger.debug("callback: on_inference_end()")
        inference_request = self._inference_requests.get(inference_request_uuid)
        if inference_request is not None:
            inference_request.on_inference_end()

            self.global_progress.inference_finished(
                current=inference_request.global_progress_current,
                total=inference_request.global_progress_total,
            )

    def _handle_error_in_async(self, inference_request_uuid: str, func, args, kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            inference_request = self._inference_requests.get(inference_request_uuid, None)
            if inference_request is not None:
                inference_request.exception = e
            logger.error(f"Error in {func.__name__} function: {e}", exc_info=True)

    def schedule_task(self, func, *args, **kwargs) -> Tuple[InferenceRequest, Future]:
        inference_request = kwargs.get("inference_request", None)
        if inference_request is None:
            inference_request = self.create()
            kwargs["inference_request"] = inference_request
        self._on_inference_start(inference_request)
        future = self._executor.submit(
            self._handle_error_in_async,
            inference_request.uuid,
            func,
            args,
            kwargs,
        )
        end_callback = partial(
            self._on_inference_end, inference_request_uuid=inference_request.uuid
        )
        future.add_done_callback(end_callback)
        logger.debug("Scheduled task.", extra={"inference_request_uuid": inference_request.uuid})
        return inference_request, future

    def run(self, func, *args, **kwargs):
        inference_request, future = self.schedule_task(func, *args, **kwargs)
        future.result()
        return inference_request.pop_pending_results()



# coding: utf-8

# docs
from __future__ import annotations

from typing import Any, Dict, List, Optional, Tuple

from supervisely.api.module_api import ApiField
from supervisely.collection.key_indexed_collection import KeyIndexedCollection
from supervisely.sly_logger import logger
from supervisely.video_annotation.frame import Frame
from supervisely.video_annotation.key_id_map import KeyIdMap
from supervisely.video_annotation.video_figure import VideoFigure
from supervisely.video_annotation.video_object_collection import VideoObjectCollection


class FrameCollection(KeyIndexedCollection):
    """
    Collection of Frame instances indexed by frame index. Immutable.

    :Usage Example:

        .. code-block:: python

            import supervisely as sly

            # Create two frames for collection
            fr_index_1 = 7
            frame_1 = sly.Frame(fr_index_1)
            fr_index_2 = 10
            frame_2 = sly.Frame(fr_index_2)

            # Create FrameCollection
            fr_collection = sly.FrameCollection([frame_1, frame_2])
            print(fr_collection.to_json())
            # Output: [
            #     {
            #         "index": 7,
            #         "figures": []
            #     },
            #     {
            #         "index": 10,
            #         "figures": []
            #     }
            # ]

            # Add item to FrameCollection
            frame_3 = sly.Frame(12)

            # Remember that TagCollection is immutable, and we need to assign new instance of TagCollection to a new variable
            new_fr_collection = fr_collection.add(frame_3)
            print(new_fr_collection.to_json())
            # Output: [
            #     {
            #         "index": 7,
            #         "figures": []
            #     },
            #     {
            #         "index": 10,
            #         "figures": []
            #     },
            #     {
            #         "index": 12,
            #         "figures": []
            #     }
            # ]

            # You can also add multiple items to collection
            frame_3 = sly.Frame(12)
            frame_4 = sly.Frame(15)

            # Remember that TagCollection is immutable, and we need to assign new instance of TagCollection to a new variable
            new_fr_collection = fr_collection.add_items([frame_3, frame_4])
            print(new_fr_collection.to_json())
            # Output: [
            #     {
            #         "index": 7,
            #         "figures": []
            #     },
            #     {
            #         "index": 10,
            #         "figures": []
            #     },
            #     {
            #         "index": 12,
            #         "figures": []
            #     },
            #     {
            #         "index": 15,
            #         "figures": []
            #     }
            # ]

            # Has key, checks if given key exist in collection
            fr_collection.has_key(7)
            # Output: True

            # Find intersection of given list of instances with collection items
            frame_1 = sly.Frame(7)
            frame_2 = sly.Frame(10)
            fr_collection = sly.FrameCollection([frame_1, frame_2])
            frame_3 = sly.Frame(12)
            frames_intersections = fr_collection.intersection([frame_3])
            print(frames_intersections.to_json())
            # Output: []

            frames_intersections = fr_collection.intersection([frame_2])
            print(frames_intersections.to_json())
            # Output: [
            #     {
            #         "index": 10,
            #         "figures": []
            #     }
            # ]

            # Note, two frames with the same index values are not equal
            frame_4 = sly.Frame(10)
            frames_intersections = fr_collection.intersection([frame_4])
            # Output:
            # ValueError: Different values for the same key 10

            # Find difference between collection and given list of Frames
            frames_difference = fr_collection.difference([frame_2])
            print(frames_difference.to_json())
            # Output: [
            #     {
            #         "index": 7,
            #         "figures": []
            #     }
            # ]

            # Merge collection and given list of FrameCollection
            frame_3 = sly.Frame(12)
            frame_4 = sly.Frame(15)
            over_collection = sly.FrameCollection([frame_3, frame_4])
            merged_collection = fr_collection.merge(over_collection)
            print(merged_collection.to_json())
            # Output: [
            #     {
            #         "index": 12,
            #         "figures": []
            #     },
            #     {
            #         "index": 15,
            #         "figures": []
            #     },
            #     {
            #         "index": 7,
            #         "figures": []
            #     },
            #     {
            #         "index": 10,
            #         "figures": []
            #     }
            # ]
    """

    item_type = Frame

    def get(self, key: str, default: Optional[Any] = None) -> Frame:
        """
        Get item from collection with given key(name) and set a default if item does not exist.

        :param key: Name of Frame in collection.
        :type key: str
        :param default: The value that is returned if there is no key in the collection.
        :type default:  Optional[Any]
        :returns: Frame, Slice or PointcloudEpisodeFrame object
        :rtype: :class:`~supervisely.collection.key_indexed_collection.KeyObject`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                frame_index = 7
                geometry = sly.Rectangle(0, 0, 100, 100)
                class_car = sly.ObjClass('car', sly.Rectangle)
                object_car = sly.VideoObject(class_car)
                figure_car = sly.VideoFigure(object_car, geometry, frame_index)

                frame = sly.Frame(frame_index, figures=[figure_car])
                frame_collection = sly.FrameCollection([frame])

                item = frame_collection.get(frame_index)
                print(item.to_json())
                # Output: {
                #     "figures": [
                #         {
                #         "geometry": {
                #             "points": {
                #             "exterior": [
                #                 [0, 0],
                #                 [100, 100]
                #             ],
                #             "interior": []
                #             }
                #         },
                #         "geometryType": "rectangle",
                #         "key": "713968a7d5384709bc5d4e63cd4535f2",
                #         "objectKey": "3342e68eff3b44dcb75712499265be55"
                #         }
                #     ],
                #     "index": 7
                # }
        """
        return super().get(key, default)

    def to_json(self, key_id_map: KeyIdMap = None) -> List[Dict]:
        """
        Convert the FrameCollection to a list of json dicts. Read more about `Supervisely format <https://docs.supervisely.com/data-organization/00_ann_format_navi>`_.

        :param key_id_map: KeyIdMap object.
        :type key_id_map: :class:`~supervisely.video_annotation.key_id_map.KeyIdMap`, optional
        :returns: List of dicts in json format
        :rtype: List[dict]

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                frame_1 = sly.Frame(7)
                frame_2 = sly.Frame(10)

                # Create FrameCollection
                fr_collection = sly.FrameCollection([frame_1, frame_2])
                print(fr_collection.to_json())
                # Output: [
                #     {
                #         "index": 7,
                #         "figures": []
                #     },
                #     {
                #         "index": 10,
                #         "figures": []
                #     }
                # ]
        """
        return [frame.to_json(key_id_map) for frame in self]

    @classmethod
    def from_json(
        cls,
        data: List[Dict],
        objects: VideoObjectCollection,
        frames_count: Optional[int] = None,
        key_id_map: Optional[KeyIdMap] = None,
        skip_corrupted: Optional[bool] = False,
    ) -> FrameCollection:
        """
        Convert a list of json dicts to FrameCollection. Read more about `Supervisely format <https://docs.supervisely.com/data-organization/00_ann_format_navi>`_.

        :param data: List with dicts in json format.
        :type data: List[dict]
        :param objects: VideoObjectCollection object.
        :type objects: :class:`~supervisely.video_annotation.video_object_collection.VideoObjectCollection`
        :param frames_count: Number of frames in video.
        :type frames_count: int, optional
        :param key_id_map: KeyIdMap object.
        :type key_id_map: :class:`~supervisely.video_annotation.key_id_map.KeyIdMap`, optional
        :returns: FrameCollection object.
        :rtype: :class:`~supervisely.video_annotation.frame_collection.FrameCollection`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                fr_collection_json = [
                    {"index": 7, "figures": []},
                    {"index": 10, "figures": []}
                ]

                objects = []
                fr_collection = sly.FrameCollection.from_json(fr_collection_json, objects)
        """
        frames = []
        for frame_json in data:
            try:
                frame = cls.item_type.from_json(frame_json, objects, frames_count, key_id_map)
                frames.append(frame)
            except Exception as e:
                if skip_corrupted:
                    logger.warning(f"Skipping corrupted frame: {e}", exc_info=True)
                    continue
                else:
                    raise e
        return cls(frames)

    def __str__(self):
        return "Frames:\n" + super(FrameCollection, self).__str__()

    @property
    def figures(self) -> List[VideoFigure]:
        """
        Get figures from all frames in collection.

        :returns: List of figures from all frames in collection
        :rtype: List[:class:`~supervisely.video_annotation.video_figure.VideoFigure`]

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                fr_index_1 = 7
                geometry = sly.Rectangle(0, 0, 100, 100)
                obj_class_car = sly.ObjClass('car', sly.Rectangle)
                video_object_car = sly.VideoObject(obj_class_car)
                video_figure_car = sly.VideoFigure(video_object_car, geometry, fr_index_1)
                frame_1 = sly.Frame(fr_index_1, figures=[video_figure_car])

                fr_index_2 = 10
                geometry = sly.Rectangle(0, 0, 500, 600)
                obj_class_bus = sly.ObjClass('bus', sly.Rectangle)
                video_object_bus = sly.VideoObject(obj_class_bus)
                video_figure_bus = sly.VideoFigure(video_object_bus, geometry, fr_index_2)
                frame_2 = sly.Frame(fr_index_2, figures=[video_figure_bus])

                fr_collection = sly.FrameCollection([frame_1, frame_2])
                figures = fr_collection.figures
        """
        figures_array = []
        for frame in self:
            figures_array.extend(frame.figures)
        return figures_array

    def get_figures_and_keys(self, key_id_map: KeyIdMap) -> Tuple[List[Dict], List[str]]:
        """
        Get figures from all frames in collection in json format, keys from all figures in frames in collection.

        :param key_id_map: KeyIdMap object.
        :type key_id_map: :class:`~supervisely.video_annotation.key_id_map.KeyIdMap`
        :returns: Figures from all frames in collection in json format, keys from all figures in frames in collection
        :rtype: Tuple[List[dict], List[str]]

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                from supervisely.video_annotation.key_id_map import KeyIdMap

                key_id_map = KeyIdMap()

                fr_index_1 = 7
                geometry = sly.Rectangle(0, 0, 100, 100)
                obj_class_car = sly.ObjClass('car', sly.Rectangle)
                video_object_car = sly.VideoObject(obj_class_car)
                video_figure_car = sly.VideoFigure(video_object_car, geometry, fr_index_1)
                frame_1 = sly.Frame(fr_index_1, figures=[video_figure_car])

                fr_index_2 = 10
                geometry = sly.Rectangle(0, 0, 500, 600)
                obj_class_bus = sly.ObjClass('bus', sly.Rectangle)
                video_object_bus = sly.VideoObject(obj_class_bus)
                video_figure_bus = sly.VideoFigure(video_object_bus, geometry, fr_index_2)
                frame_2 = sly.Frame(fr_index_2, figures=[video_figure_bus])

                fr_collection = sly.FrameCollection([frame_1, frame_2])
                figures, keys = fr_collection.get_figures_and_keys(key_id_map)
                print(keys) # [UUID('0ac041b2-314e-4f6b-9d38-704b341fb383'), UUID('88aa1cb3-b1e3-480f-8ace-6346c9a9daba')]

                print(figures)
                # Output: [
                #     {
                #         "key": "a8cae05d6b8c4a67b18004130941fdec",
                #         "objectKey": "cc9a9475d360481c9753f8ac3c63f8b7",
                #         "geometryType": "rectangle",
                #         "geometry": {
                #             "points": {
                #                 "exterior": [
                #                     [
                #                         0,
                #                         0
                #                     ],
                #                     [
                #                         100,
                #                         100
                #                     ]
                #                 ],
                #                 "interior": []
                #             }
                #         },
                #         "meta": {
                #             "frame": 7
                #         }
                #     },
                #     {
                #         "key": "6e00287acc4644dfb21d67406534080b",
                #         "objectKey": "cad78d53ffc84e69a28f5f8941be9021",
                #         "geometryType": "rectangle",
                #         "geometry": {
                #             "points": {
                #                 "exterior": [
                #                     [
                #                         0,
                #                         0
                #                     ],
                #                     [
                #                         600,
                #                         500
                #                     ]
                #                 ],
                #                 "interior": []
                #             }
                #         },
                #         "meta": {
                #             "frame": 10
                #         }
                #     }
                # ]
        """
        keys = []
        figures_json = []
        for frame in self:
            for figure in frame.figures:
                keys.append(figure.key())
                figure_json = figure.to_json(key_id_map)
                figure_json[ApiField.META] = figure.get_meta()
                figures_json.append(figure_json)
        return figures_json, keys
